function BlockMove(event) {
  // Tell the browser to not to "bounce" the content.
  event.preventDefault();
}

function parseDate(date) {
  // date: "02.01.2020"
  // start, end: [month, day, year]
  var strs = date.split('.');
  var [month, day, year] = [parseInt(strs[1]), parseInt(strs[0]), parseInt(strs[2])];
  return [month, day, year];
}

function updateDateFrame(date) {
  // date format [yy, mm, dd]
  var [yy, mm, dd] = date;
  var [syy, smm, sdd] = startdate;
  var [eyy, emm, edd] = enddate;
  if (syy == -1 || eyy == -1) {
    startdate[0] = yy;
    startdate[1] = mm;
    startdate[2] = dd;
    enddate[0] = yy;
    enddate[1] = mm;
    enddate[2] = dd;
    return;
  }
  if (yy < syy || yy == syy && mm < smm || yy == syy && mm == smm && dd < sdd) {
    startdate[0] = yy;
    startdate[1] = mm;
    startdate[2] = dd;
  }
  if (yy > eyy || yy == eyy && mm > emm || yy == eyy && mm == emm && dd > edd) {
    enddate[0] = yy;
    enddate[1] = mm;
    enddate[2] = dd;
  }
}

function updateRenderDate() {
  renderDate[0] = startdate[0];
  renderDate[1] = startdate[1];
  renderDate[2] = startdate[2];
}

function initDataMatrix() {
  // Create a table has the following format
  // Each block represents a day
  //      | Jan   | Feb   | Mar   | ...   | Dec   |
  // 2019 | | | | | | | | | | | | | | | | | | | | |
  // 2020 | | | | | | | | | | | | | | | | | | | | |
  for (var year = startyear; year <= endyear; year++) {
    indivMatrix.push([]);
    indivWhMatrix.push([]);
    totalMatrix.push([]);
    // NOTE: create month grids
    for (var mm = 0; mm < 12; mm++) {
      var yy = year - startyear;
      indivMatrix[yy].push([]);
      indivWhMatrix[yy].push([]);
      totalMatrix[yy].push([]);
      // NOTE: create day grids
      var numdays = mdays[mm];
      if (year % 4 == 0 && mm == 2 - 1) numdays++;
      for (var dd = 0; dd < numdays; dd++) {
        indivMatrix[yy][mm].push(new DailyCases([mm + 1, dd + 1, year]));
        indivWhMatrix[yy][mm].push(new DailyCases([mm + 1, dd + 1, year]));
      }
    }
  }

}

function loadIndivDataset() {
  var table = indivDataset;
  var nRows = table.getRowCount();
  for (var row = 0; row < nRows; row++) {
    try { var age = table.getNum(row, headers.age); } catch { var age = -1 };
    var sex = table.getString(row, headers.sex);
    var city = table.getString(row, headers.city);
    var province = table.getString(row, headers.province);
    var country = table.getString(row, headers.country);
    try { var lat = table.getNum(row, headers.lat); } catch { var lat = NaN };
    try { var lng = table.getNum(row, headers.lng); } catch { var lng = NaN };
    var dateSympStr = table.getString(row, headers.dateSymp);
    var dateConfStr = table.getString(row, headers.dateConfirmed);
    var travHistStr = table.getString(row, headers.travHist);
    var travLoca = table.getString(row, headers.travLoca);

    // Get the corresponding daily data cell
    var [mm, dd, yy] = parseDate(dateConfStr);
    if (!mm || !dd || !yy || !lat || !lng)
      continue;

    var vcase = new VirusCase(sex, age, dateSympStr, dateConfStr, travHistStr, travLoca);
    var dailyCases = indivMatrix[yyidx(yy)][mmidx(mm)][ddidx(dd)];
    dailyCases.add_case(lat, lng, city, province, country, vcase);
    updateDateFrame([yy, mm, dd])
  }
  print("indivMatrix loaded!");
  // DEBUG: print out the keys in each dailycases
  // for(var year = startyear; year <= endyear; year++) {
  //   for(var mm = 0; mm < 12; mm++) {
  //     for (var dd = 0; dd < 31; dd++) {
  //       var yy = year - startyear;
  //       dailyCases = indivMatrix[yy][mm][dd];
  //       print("daily cases: ", Object.keys(dailyCases.areas));
  //     }
  //   }
  // } 
}

function loadIndivWhDataset() {
  var table = indivWhDataset;
  var nRows = table.getRowCount();
  for (var row = 0; row < nRows; row++) {
    try { var age = table.getNum(row, headers.age); } catch { var age = -1 };
    var sex = table.getString(row, headers.sex);
    var city = table.getString(row, headers.city);
    var province = table.getString(row, headers.province);
    var country = table.getString(row, headers.country);
    try { var lat = table.getNum(row, headers.lat); } catch { var lat = NaN };
    try { var lng = table.getNum(row, headers.lng); } catch { var lng = NaN };
    var dateSympStr = table.getString(row, headers.dateSymp);
    var dateConfStr = table.getString(row, headers.dateConfirmed);
    var travHistStr = table.getString(row, headers.travHist);
    var travLoca = table.getString(row, headers.travLoca);

    // Get the corresponding daily data cell
    var [mm, dd, yy] = parseDate(dateConfStr);
    if (!mm || !dd || !yy || !lat || !lng)
      continue;
    var vcase = new VirusCase(sex, age, dateSympStr, dateConfStr, travHistStr, travLoca);
    var dailyCases = indivWhMatrix[yyidx(yy)][mmidx(mm)][ddidx(dd)];
    dailyCases.add_case(lat, lng, city, province, country, vcase);
    updateDateFrame([yy, mm, dd])
  }
  print("indivWhMatrix loaded!");
  // DEBUG: print out the keys in each dailycases
  // for(var year = startyear; year <= endyear; year++) {
  //   for(var mm = 0; mm < 12; mm++) {
  //     for (var dd = 0; dd < 31; dd++) {
  //       var yy = year - startyear;
  //       dailyCases = indivWhMatrix[yy][mm][dd];
  //       print("daily cases: ", Object.keys(dailyCases.areas));
  //     }
  //   }
  // } 
}

function getCopiedTotalStats(total) {
  var copy = {};
  for(var key in total) {
    stats = total[key];
    copiedStats = new AreaStats(stats.lat, stats.lng);
    copiedStats.maleNum = stats.maleNum;
    copiedStats.femaleNum = stats.femaleNum;
    copiedStats.totalNum = stats.totalNum;
    copy[key] = copiedStats;
  }
  return copy;
}

function updateTotalCases() {
  var total = {};
  var yy;
  for(var year = startyear; year <= endyear; year++) {
    for(var mm = 0; mm < 12; mm++) {
      var days = mdays[mm];
      if (mm == 1 && year % 4 == 0) days += 1;
      for(var dd = 0; dd < days; dd++) {
        yy = year - startyear;
        dailyCases = indivMatrix[yy][mm][dd];
        // update total status by adding up the dailyCases outside wuhan
        for(var key in dailyCases.areas) {
          areaCases = dailyCases.areas[key];
          areaStats = total[key];
          if(!areaStats) {
            areaStats = new AreaStats(areaCases.lat, areaCases.lng);
            total[key] = areaStats;
          }
          areaStats.maleNum += areaCases.maleCases.length;
          areaStats.femaleNum += areaCases.femaleCases.length;
          areaStats.totalNum += areaCases.maleCases.length + areaCases.femaleCases.length + areaCases.nonsexCases.length;
        }
        dailyCases = indivWhMatrix[yy][mm][dd];
        // update total status by adding up the dailyCases outside wuhan
        for(var key in dailyCases.areas) {
          areaCases = dailyCases.areas[key];
          areaStats = total[key];
          if(!areaStats) {
            areaStats = new AreaStats(areaCases.lat, areaCases.lng);
            total[key] = areaStats;
          }
          areaStats.maleNum += areaCases.maleCases.length;
          areaStats.femaleNum += areaCases.femaleCases.length;
          areaStats.totalNum += areaCases.maleCases.length + areaCases.femaleCases.length + areaCases.nonsexCases.length;
        }
        // Copy the current total area stats into the totalMatrix
        totalMatrix[yy][mm].push(getCopiedTotalStats(total));
      }
    }
  }
}

function updateMesh() {
  totalMesh = createMesh();
  maleMash = createMesh();
  femaleMash = createMesh();
  var xidx, yidx;
  
  // print("Total status...");
  var [yy, mm, dd] = renderDate;
  var areas = totalMatrix[yyidx(yy)][mmidx(mm)][ddidx(dd)];
  for(var key in areas) {
    stats = areas[key];
    xidx = Math.floor(stats.lng / div) + 180 / div;
    yidx = Math.floor(stats.lat / div) + 90 / div;
    // totalMesh[yidx][xidx] += stats.totalNum;
    convSum(totalMesh, xidx, yidx, stats.totalNum, 4);
    convSum(maleMash, xidx, yidx, stats.maleNum, 4);
    convSum(femaleMash, xidx, yidx, stats.femaleNum, 4);
  }
}

function getTotalDays() {
  const [syy, smm, sdd] = startdate;
  const [eyy, emm, edd] = enddate;
  var start = new Date(syy, smm, sdd);
  var end = new Date(eyy, emm, edd);
  var deltaTime = end.getTime() - start.getTime();
  return Math.floor(deltaTime / (1000 * 3600 * 24) + 1);
}

function updateDate() {
  var days = renderDateSlider.value();
  print(startdate);
  var [yy, mm, dd] = startdate;
  if (mm > 0) mm -= 1;
  var date = new Date(yy, mm, dd);
  date.setTime(date.getTime() + days * 1000 * 3600 * 24);

  renderDate[0] = date.getFullYear();
  renderDate[1] = date.getMonth() + 1;
  renderDate[2] = date.getDate();
  labelDateValue.html('' + renderDate);
}